import astropy.units as u

from .base import Backend


class GalaBackend(Backend):
    def __imports__():
        from gala import dynamics, potential, units

    @property
    def ORBIT_TYPE(self):
        return self.dynamics.orbit.Orbit

    def _compute_orbit(
        self,
        skycoord,
        pot,
        dt,
        steps,
        pattern_speed=0 * u.km / u.s / u.kpc,
        integrator=None,
        integrator_kwargs=None,
        cython_if_possible=True,
        store_all=True,
        **time_spec,
    ):
        pos = [skycoord.x.value, skycoord.y.value, skycoord.z.value] * skycoord.x.unit
        vel = [
            skycoord.v_x.value,
            skycoord.v_y.value,
            skycoord.v_z.value,
        ] * skycoord.v_x.unit
        ics = self.dynamics.PhaseSpacePosition(pos, vel)

        if integrator_kwargs is None:
            integrator_kwargs = {}
        orbit = pot.integrate_orbit(
            ics,
            integrator,
            integrator_kwargs,
            cython_if_possible,
            store_all,
            dt=dt,
            n_steps=steps - 1,
        )

        # In the case of a single element collection of initial conditions,
        # the orbit is returned as a 1D array. We need to reshape it to the
        # expected shape.
        expected_shape = (steps, *skycoord.shape)
        if orbit.shape != expected_shape:
            orbit = orbit.reshape(expected_shape)
        return orbit

    def _extract_points(self, orbit, pattern_speed=0 * u.km / u.s / u.kpc):
        frame = self.potential.ConstantRotatingFrame(
            Omega=[0, 0, pattern_speed.value] * pattern_speed.unit,
            units=self.units.galactic,
        )
        orbit = orbit.to_frame(frame)
        return orbit.data.T



import astropy.coordinates as coord
import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import CartesianDifferential
from gala import integrate as gi
from gala import potential as gp
from gala import units as gu

from pidgey import GalaBackend, get_backend_from


@pytest.fixture
def backend():
    return GalaBackend()


@pytest.fixture
def potential():
    return gp.NFWPotential.from_circular_velocity(
        v_c=200 * u.km / u.s, r_s=10.0 * u.kpc, units=gu.galactic
    )


def test_detection(potential):
    backend = get_backend_from(potential)
    assert isinstance(backend, GalaBackend)


def test_single_orbit(potential, backend):
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(c, potential, dt=0.1 * u.Gyr, steps=STEPS)
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


def test_single_orbit_with_alternate_integrator(potential, backend):
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(
        c,
        potential,
        dt=0.1 * u.Gyr,
        steps=STEPS,
        integrator=gi.LeapfrogIntegrator,
    )
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


def test_single_orbit_with_pattern_speed(potential, backend):
    # NOTE: potential is not rotating, only tests that the pattern speed is accepted by the backend
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(
        c,
        potential,
        dt=0.1 * u.Gyr,
        steps=STEPS,
        pattern_speed=50 * u.km / u.s / u.kpc,
    )
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


@pytest.mark.parametrize("orbits", [1, 10])
def test_multiple_orbits(potential, backend, orbits):
    ORBITS = 10
    STEPS = 50

    xs = np.ones(ORBITS) * u.kpc
    ys = np.ones(ORBITS) * u.kpc
    zs = np.ones(ORBITS) * u.kpc
    v_xs = np.zeros(ORBITS) * u.km / u.s
    v_ys = np.linspace(100, 300, ORBITS) * u.km / u.s
    v_zs = np.zeros(ORBITS) * u.km / u.s
    c = coord.SkyCoord(
        x=xs,
        y=ys,
        z=zs,
        v_x=v_xs,
        v_y=v_ys,
        v_z=v_zs,
        frame="galactocentric",
        representation_type="cartesian",
    )

    orbit = backend.compute_orbit(c, potential, dt=0.1 * u.Gyr, steps=STEPS)
    print(orbit.T)
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == ORBITS
    assert len(orbit[0]) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == ORBITS
    assert len(orbit[0]) == STEPS



from abc import abstractmethod

import astropy.coordinates as coord
import astropy.units as u
from iext import ExtendImports


class Backend(ExtendImports):
    """
    Base class for orbit integration backends.

    This class defines the interface for orbit integration through
    two methods: `compute_orbit` and `get_points`. Two abstract
    methods, `_compute_orbit` and `_extract_points`, and an abstract
    attribute, `ORBIT_TYPE`, must be implemented to define a new backend.

    Attributes:
        ORBIT_TYPE (type): The type of the orbit object returned by the backend.
    """

    def __imports__():
        ...

    def __init__(self):
        self._args = None
        self._result = None

    @property
    @abstractmethod
    def ORBIT_TYPE(self):
        """
        The type of the orbit object returned by the backend.
        """
        pass

    @abstractmethod
    def _compute_orbit(
        self,
        skycoord,
        pot,
        dt,
        steps,
        pattern_speed=0 * u.km / u.s / u.kpc,
        **integration_kwargs,
    ):
        pass

    def compute_orbit(
        self,
        skycoord,
        pot,
        dt,
        steps,
        pattern_speed=0 * u.km / u.s / u.kpc,
        **integration_kwargs,
    ):
        """
        Compute an orbit for a given initial condition (through a SkyCoord object),
        through a given potential.

        Args:
            skycoord (astropy.coordinates.SkyCoord): The initial conditions of the orbit.
            pot (object): The potential to compute the orbit in.
            dt (astropy.units.Quantity): The time step for the integration.
            steps (int): The number of steps to integrate over.
            pattern_speed (astropy.units.Quantity): The pattern speed of the potential.
            **integration_kwargs: Additional arguments for integration routines.

        Returns:
            astropy.coordinates.SkyCoord: The points from the orbit integration.
        """
        if not isinstance(skycoord, coord.SkyCoord):
            raise TypeError(
                "coord must be passed in as a astropy.coordinates.SkyCoord object."
                f"{skycoord} is not a astropy.coordinates.SkyCoord object."
            )
        skycoord.representation_type = "cartesian"
        if not (
            hasattr(skycoord, "v_x")
            and hasattr(skycoord, "v_y")
            and hasattr(skycoord, "v_z")
        ):
            raise TypeError(
                "coord must have 3-D velocity information to compute an orbit."
            )
        if not isinstance(dt, u.Quantity):
            raise TypeError(
                "dt must be passed in as a astropy.units.Quantity object."
                f"{dt} is not a astropy.coordinates.SkyCoord object."
            )
        self._args = skycoord, pot, dt, steps
        self._result = self._compute_orbit(
            skycoord, pot, dt, steps, pattern_speed, **integration_kwargs
        )
        return self._extract_points(self._result, pattern_speed)

    @abstractmethod
    def _extract_points(self, orbit, pattern_speed=0 * u.km / u.s / u.kpc):
        pass

    def get_points(self, pattern_speed=0 * u.km / u.s / u.kpc):
        """
        Get the points from the orbit integration result in a SkyCoord object.

        Args:
            pattern_speed (astropy.units.Quantity): The pattern speed of the potential.

        Returns:
            astropy.coordinates.SkyCoord: The points from the orbit integration.
        """
        if not isinstance(self._result, self.ORBIT_TYPE):
            raise TypeError("No computed orbit to retrive points from.")
        return self._extract_points(self._result, pattern_speed)



import astropy.coordinates as coord
import astropy.units as u
import galpy
import numpy as np
import pytest
from astropy.coordinates import CartesianDifferential

from pidgey import GalpyBackend, get_backend_from


@pytest.fixture
def backend():
    return GalpyBackend()


@pytest.fixture
def potential():
    return galpy.potential.MWPotential2014


def test_detection(potential):
    backend = get_backend_from(potential)
    assert isinstance(backend, GalpyBackend)


def test_single_orbit(potential, backend):
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(c, potential, dt=0.1 * u.Gyr, steps=STEPS)
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


def test_single_orbit_with_alternate_integrator(potential, backend):
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(
        c, potential, dt=0.1 * u.Gyr, steps=STEPS, method="leapfrog_c"
    )
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


def test_single_orbit_with_pattern_speed(potential, backend):
    # NOTE: potential is not rotating, only tests that the pattern speed is accepted by the backend
    STEPS = 50
    c = coord.SkyCoord(
        ra=20.0 * u.deg,
        dec=30.0 * u.deg,
        distance=2.0 * u.kpc,
        pm_ra_cosdec=-10.0 * u.mas / u.yr,
        pm_dec=20.0 * u.mas / u.yr,
        radial_velocity=50.0 * u.km / u.s,
        galcen_distance=8.0 * u.kpc,
        z_sun=15.0 * u.pc,
        galcen_v_sun=CartesianDifferential([10.0, 235.0, 7.0] * u.km / u.s),
    )
    orbit = backend.compute_orbit(
        c,
        potential,
        dt=0.1 * u.Gyr,
        steps=STEPS,
        pattern_speed=42.0 * u.km / u.s / u.kpc,
    )
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS
    orbit = backend.get_points()
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == STEPS


@pytest.mark.parametrize("orbits", [1, 10])
def test_multiple_orbits(potential, backend, orbits):
    ORBITS = 10
    STEPS = 50

    xs = np.ones(ORBITS) * u.kpc
    ys = np.ones(ORBITS) * u.kpc
    zs = np.ones(ORBITS) * u.kpc
    v_xs = np.zeros(ORBITS) * u.km / u.s
    v_ys = np.linspace(100, 300, ORBITS) * u.km / u.s
    v_zs = np.zeros(ORBITS) * u.km / u.s
    c = coord.SkyCoord(
        x=xs,
        y=ys,
        z=zs,
        v_x=v_xs,
        v_y=v_ys,
        v_z=v_zs,
        frame="galactocentric",
        representation_type="cartesian",
    )

    orbit = backend.compute_orbit(c, potential, dt=0.1 * u.Gyr, steps=STEPS)
    assert isinstance(orbit, coord.representation.CartesianRepresentation)
    assert len(orbit) == ORBITS
    assert len(orbit[0]) == STEPS



"""Python interface for dynamics in galaxies, supporting multiple backends."""

__all__ = [
    "get_backend_from",
    # Backends
    "AgamaBackend",
    "GalaBackend",
    "GalpyBackend",
]

import inspect as _inspect
from collections.abc import Sequence as _Sequence

from .agama_backend import AgamaBackend
from .gala_backend import GalaBackend
from .galpy_backend import GalpyBackend


def get_backend_from(obj):
    """
    Get the backend associated with an object's module of origin.

    Args:
        obj (object): An object whose module is used to determine the appropriate backend.

    Returns:
        Backend: The backend associated with the object's module.
    """
    if isinstance(obj, _Sequence) and len(obj) > 0:
        obj = obj[0]
    module = _inspect.getmodule(obj) or _inspect.getmodule(type(obj))
    if module is None:
        return None

    pkg, *_ = module.__name__.partition(".")
    if pkg == "agama":
        return AgamaBackend()
    elif pkg == "gala":
        return GalaBackend()
    elif pkg == "galpy":
        return GalpyBackend()
    return None



import astropy.coordinates as coord
import astropy.units as u
import numpy as np

from .base import Backend


class AgamaBackend(Backend):
    def __imports__():
        import agama

        agama.setUnits(mass=1, length=1, velocity=1)

    @property
    def ORBIT_TYPE(self):
        return np.ndarray

    def _compute_orbit(
        self,
        skycoord,
        pot,
        dt,
        steps,
        pattern_speed=0 * u.km / u.s / u.kpc,
        **integration_kwargs,
    ):
        pos = [getattr(skycoord, attr).to(u.kpc).value for attr in ("x", "y", "z")]
        vel = [
            getattr(skycoord, attr).to(u.km / u.s).value
            for attr in ("v_x", "v_y", "v_z")
        ]
        posvel = np.array([*pos, *vel]).T
        orbit = self.agama.orbit(
            potential=pot,
            ic=posvel,
            time=(dt * steps).to(u.Gyr).value,
            trajsize=steps,
            Omega=pattern_speed.to(u.km / u.s / u.kpc).value,
            **integration_kwargs,
        )
        return orbit

    def _extract_points(self, orbit, pattern_speed=0 * u.km / u.s / u.kpc):
        skycoord, pot, dt, steps = self._args
        if not skycoord.shape:
            orbit = [orbit]
        xs = []
        ys = []
        zs = []
        for times, posvel in orbit:
            x, y, z, *_ = posvel.T
            xs.append(x)
            ys.append(y)
            zs.append(z)
        if not skycoord.shape:
            (xs,) = xs
            (ys,) = ys
            (zs,) = zs
        return coord.representation.CartesianRepresentation(xs, ys, zs)



import astropy.coordinates as coord
import astropy.units as u
import numpy as np

from .base import Backend


class GalpyBackend(Backend):
    def __imports__():
        from galpy import orbit
        from galpy.util import conversion

    @property
    def ORBIT_TYPE(self):
        return self.orbit.Orbit

    def _compute_orbit(
        self,
        skycoord,
        pot,
        dt,
        steps,
        pattern_speed=0 * u.km / u.s / u.kpc,
        **integration_kwargs,
    ):
        # unit consistency issues... look into this later
        pot_units = self.conversion.get_physical(pot)
        orbit = self.orbit.Orbit(skycoord, **pot_units)

        t = np.arange(steps) * dt
        orbit.integrate(t, pot, **integration_kwargs)
        return orbit

    def _extract_points(self, orbit, pattern_speed=0 * u.km / u.s / u.kpc):
        skycoord, pot, dt, steps = self._args
        t = np.arange(steps) * dt
        phi_offset = t * pattern_speed

        R = orbit.R(t)
        phi = orbit.phi(t) + phi_offset.to(u.dimensionless_unscaled)

        x = R * np.cos(phi.value)
        y = R * np.sin(phi.value)
        z = orbit.z(t)

        return coord.representation.CartesianRepresentation(x, y, z)
